__all__ = ('DuckOperator',)

from pathlib import Path

import duckdb
import duckdb_engine
from aloha.logger import LOG
from sqlalchemy import create_engine, text

LOG.debug('duckdb version = %s, duckdb_engine = %s ' % (duckdb.__version__, duckdb_engine.__version__))


class DuckOperator:
    def __init__(self, db_config, **kwargs):
        """db_config example:
        {
            "path": "/path/to/db.duckdb",     # file path of duckdb, use ":memory:" for in-memory mode
            "schema": "sales",                # optional, 'main' by default
            "read_only": True,                # optional, False by default, (will set to False if in in-memory mode)
            "config": {"memory_limit": "500mb"}, # optional, duckdb connection configs
        }
        """
        self._config = {
            'path': db_config.get('path', ':memory:'),
            'schema': db_config.get('schema', 'main'),
            'read_only': bool(db_config.get('read_only', False)),
            'config': db_config.get('config', {}),
            'auto_commit': db_config.get('auto_commit', True),
        }

        if not self._config['path'] or self._config['path'] == ':memory:':  # in-memory mode
            self._config['path'] = ':memory:'

            if self._config['read_only']:  # in-memory mode cannot be read-only
                LOG.warning("In-memory database cannot be read-only. Setting read_only=False.")
                self._config['read_only'] = False

        else:
            self._prepare_database()

        try:
            str_connection = f"duckdb:///{self._config['path']}"
            self.engine = create_engine(
                str_connection,
                connect_args={
                    'read_only': self._config['read_only'],
                    'config': self._config['config']
                },
                **kwargs
            )

            self._initialize_schema()
            msg = f"DuckDB connected: {self._config['path']} [schema={self._config['schema']}, read_only={self._config['read_only']}]"
            LOG.debug(msg)
        except Exception as e:
            LOG.exception(e)
            raise RuntimeError('Failed to connect to DuckDB')

    def _prepare_database(self):
        """Prepare the database file and its parent directory."""
        path = self._config['path']
        path_obj = Path(path)

        parent_dir = path_obj.parent
        if not parent_dir.exists():
            if self._config['read_only']:
                raise RuntimeError(
                    f"Directory '{parent_dir}' does not exist and read_only=True"
                )
            try:
                parent_dir.mkdir(parents=True, exist_ok=True)
                LOG.debug(f"Created directory: {parent_dir}")
            except Exception as e:
                raise RuntimeError(f"Failed to create directory '{parent_dir}': {e}")

        if not path_obj.exists():
            if self._config['read_only']:
                raise RuntimeError(
                    f"DuckDB file '{path}' does not exist and read_only=True"
                )
            try:
                LOG.debug(f"Database file not found, creating: {path}")
                duckdb.connect(path).close()
            except Exception as e:
                raise RuntimeError(f"Failed to create database file '{path}': {e}")

    def _initialize_schema(self):
        if self._config['schema'] == 'main':
            return

        try:
            if self._config['read_only']:
                result = self.engine.connext().execute(
                    text("SELECT schema_name FROM information_schema.schemata WHERE schema_name = :schema"),
                    {'schema': self._config['schema']}
                )
                if not result.fetchone():
                    raise RuntimeError(
                        f"Schema '{self._config['schema']}' does not exist and read_only=True"
                    )
            else:
                self.engine.connect().execute(text(f"CREATE SCHEMA IF NOT EXISTS {self._config['schema']}"))

            self.engine.connect().execute(text(f"SET schema '{self._config['schema']}'"))
        except Exception as e:
            raise RuntimeError(f'Failed to initialize schema: {e}')

    @property
    def connection(self):
        return self.engine

    conn = connection

    def execute_query(self, sql, *args, **kwargs):
        with self.engine.connect() as conn:
            cur = conn.execute(text(sql), *args, *kwargs)
            if self._config.get('auto_commit', True):
                conn.commit()
            return cur

    @property
    def connection_str(self) -> str:
        return f"duckdb:///{self._config['path']} [schema={self._config['schema']}, read_only={self._config['read_only']}]"



__all__ = ('get_sys_info',)

import platform
from datetime import datetime

import psutil

from ..logger import LOG

LOG.debug('Using psutil == %s' % psutil.__version__)


def get_size(bytes, suffix="B"):
    """Scale bytes to its proper format
    e.g:
        1253656 => '1.20MB'
        1253656678 => '1.17GB'
    """
    factor = 1024
    for unit in ["", "K", "M", "G", "T", "P"]:
        if bytes < factor:
            return f"{bytes:.2f}{unit}{suffix}"
        bytes /= factor


def get_os_info(*args, **kwargs) -> dict:
    ret = {}

    boot_time = datetime.fromtimestamp(psutil.boot_time())
    ret['boot_time'] = boot_time.strftime('%Y-%m-%d %H:%M:%S.%f')

    uname = platform.uname()
    ret.update(uname._asdict())
    return ret


def get_cpu_info(*args, **kwargs) -> dict:
    cpu_freq = psutil.cpu_freq()  # CPU frequencies
    ret = {
        "num_cores_physical": psutil.cpu_count(logical=False),
        "num_cores_total": psutil.cpu_count(logical=True),
        "freq_max_mhz": f"{cpu_freq.max:.2f}",
        "freq_min_mhz": f"{cpu_freq.min:.2f}",
        "freq_cur_mhz": f"{cpu_freq.current:.2f}",
        "cpu_percent_total": f"{psutil.cpu_percent()}%",
    }
    for i, percentage in enumerate(psutil.cpu_percent(percpu=True, interval=1)):
        ret['cpu_percent_core_%02d' % i] = f"{percentage}%"

    return ret


def get_mem_info(*args, **kwargs) -> dict:
    svmem = psutil.virtual_memory()
    swap = psutil.swap_memory()

    return {
        "vm_total": f"{get_size(svmem.total)}",
        "vm_available": f"{get_size(svmem.available)}",
        "vm_used": f"{get_size(svmem.used)}",
        "vm_percent": f"{svmem.percent}%",

        "swap_total": f"{get_size(swap.total)}",
        "swap_free": f"{get_size(swap.free)}",
        "swap_used": f"{get_size(swap.used)}",
        "swap_percent": f"{swap.percent}%",
    }


def get_disk_info(*args, **kwargs) -> dict:
    # get IO statistics since boot
    disk_io = psutil.disk_io_counters()
    partitions = psutil.disk_partitions()

    ret = {
        "io_total_read": f"{get_size(disk_io.read_bytes)}",
        "io_total_write": f"{get_size(disk_io.write_bytes)}",
        "partitions": []
    }

    for partition in partitions:
        part = {
            "device": f"{partition.device}",
            "mount_point": f"{partition.mountpoint}",
            "fs_type": f"{partition.fstype}",
        }

        try:
            partition_usage = psutil.disk_usage(partition.mountpoint)
            part.update({
                "size_total": f"{get_size(partition_usage.total)}",
                "size_used": f"{get_size(partition_usage.used)}",
                "size_free": f"{get_size(partition_usage.free)}",
                "percent_used": f"{partition_usage.percent}%",
            })
        except PermissionError:
            pass  # this can be caught due to the disk that isn't ready

        ret["partitions"].append(part)

    return ret


def get_net_info(*args, **kwargs) -> dict:
    # get IO statistics since boot
    net_io = psutil.net_io_counters()

    # get all network interfaces (virtual and physical)
    if_addresses = psutil.net_if_addrs()

    ret = {
        "net_total_sent": f"{get_size(net_io.bytes_sent)}",
        "net_total_received": f"{get_size(net_io.bytes_recv)}",
        "interfaces": []
    }

    for interface_name, interface_addresses in if_addresses.items():
        interface = {"name": interface_name}

        for address in interface_addresses:
            family = str(address.family).split('.')[-1]
            family = {'AF_LINK': 'mac', 'AF_INET': 'ipv4', 'AF_INET6': 'ipv6'}.get(family, family)

            interface['%s_address' % family] = address.address
            interface['%s_netmask' % family] = address.netmask
            interface['%s_broadcast' % family] = address.broadcast

        ret['interfaces'].append(interface)

    return ret


def get_sys_info(*args, **kwargs) -> dict:
    return {
        "os_info": get_os_info(),
        "cpu_info": get_cpu_info(),
        "mem_info": get_mem_info(),
        "disk_info": get_disk_info(),
        "net_info": get_net_info(),
    }


def main():
    data = get_sys_info()
    import json
    data = json.dumps(data, ensure_ascii=False, indent=2)
    print(data)


if __name__ == '__main__':
    main()



__all__ = ('APIHandler', 'APICaller', 'sign_data', 'sign_check')

import json
import logging
import uuid
from abc import ABC

from ..http import AbstractApiClient, AbstractApiHandler
from ...encrypt.hash import get_md5_of_str, get_sha256_of_str
from ...settings import SETTINGS

APP_ID_KEYS = SETTINGS.config.get('APP_ID_KEYS', {})
APP_OPTIONS = SETTINGS.config.get('APP_OPTIONS', {})
FUNC_SIGN_CHECK = {'md5': get_md5_of_str, 'sha256': get_sha256_of_str}
func_sign_check_default = FUNC_SIGN_CHECK.get(APP_OPTIONS.get('sign_method', 'md5'))


class APIHandler(AbstractApiHandler, ABC):
    MAP_ERROR_INFO = {
        'BAD_REQUEST': {'code': '5101', 'message': ['Bad request: fail to parse body as JSON object!']},
        'MISSING_ARGS': {'code': '5102', 'message': ['Required argument field(s) missing...']},
        'SIGN_CHECK_FAIL': {'code': '5104', 'message': ['Invalid sign, sign check failed!']},
    }

    async def post(self):
        body_arguments = self.request_body

        try:
            salt_uuid = body_arguments.pop('salt_uuid')
            app_id = body_arguments.pop('app_id')
            sign = body_arguments.pop('sign')
            data = body_arguments.pop('data')
        except KeyError:  # cannot find default key from parsed body
            return self.finish(self.MAP_ERROR_INFO['MISSING_ARGS'])

        is_valid_req = sign_check(salt_uuid=salt_uuid, app_id=app_id, sign=sign, data=data)  # , sign_method='sha256'
        if not is_valid_req:
            return self.finish(self.MAP_ERROR_INFO['SIGN_CHECK_FAIL'])

        resp = dict(code=5200, message=['success'])
        try:
            result = self.response(**data)  # this call may throw TypeError when argument missing
            resp['data'] = result
            resp['salt_uuid'] = salt_uuid
        except Exception as e:
            if self.LOG.level == logging.DEBUG:
                self.LOG.error(e, exc_info=True)
            return self.finish({'code': 5201, 'message': [repr(e)]})

        resp = json.dumps(resp, ensure_ascii=False, default=str, separators=(',', ':'))
        return self.finish(resp)


class APICaller(AbstractApiClient):
    APP_ID_KEYS = AbstractApiClient.config.get('APP_ID_KEYS', {})

    def wrap_request_data(
            self, data, app_id: str = None, app_key: str = None, salt_uuid: str = None, sign: str = None, sign_method: str = None
    ):
        if app_id is None:
            # if len(APP_ID_KEYS) != 1:
            #     raise RuntimeError('Please specify 1 and only 1 in APP_ID_KEYS in configurations!')
            app_id = list(self.APP_ID_KEYS.keys())[0]
        salt_uuid = salt_uuid or str(uuid.uuid1())
        sign = sign or sign_data(
            salt_uuid=salt_uuid,
            app_id=app_id,
            app_key=app_key or self.APP_ID_KEYS.get(app_id),
            data=data,
            sign_method=sign_method
        )
        return {
            'salt_uuid': salt_uuid,
            'app_id': app_id,
            'sign': sign,
            'data': data
        }


def sign_data(salt_uuid: str, app_id: str, app_key: str, data, sign_method: str = None):
    data_str = str(json.dumps(data, ensure_ascii=False, sort_keys=True, separators=(',', ':')))
    public_key = app_id + salt_uuid + data_str + app_key

    func_sign_check = func_sign_check_default if sign_method is None else FUNC_SIGN_CHECK.get(sign_method)
    if func_sign_check is None:
        raise ValueError('Invalid `sign_method`: %s' % sign_method)
    sign = func_sign_check(public_key)
    return sign


def sign_check(salt_uuid: str, app_id: str, sign: str, data, sign_method: str = None, date_time=None):
    """Sign Validation
    :param salt_uuid: Universal Unified ID for 1) Signature, 2) Log tracing
    :param app_id: APP ID
    :param sign: sing = hash(app_id + salt_uuid + data + app_key)
    :param data: data object, will be serialized to JSON string
    :param date_time: not used for now
    :param sign_method: Sign method, one of the following: md5, sha256
    :return: If the signature passed validation
    """

    func_sign_check = func_sign_check_default if sign_method is None else FUNC_SIGN_CHECK.get(sign_method)
    if func_sign_check is None:
        raise ValueError('Invalid `sign_method`: %s' % sign_method)

    app_key = APP_ID_KEYS.get(app_id)
    if app_key is None:  # APP_ID not in the dict, unknown APP_ID
        return False

    data_str = str(json.dumps(data, ensure_ascii=False, sort_keys=True, separators=(',', ':')))

    # --> Compatible with older version API
    right_sign = func_sign_check(app_id + salt_uuid + app_key)
    if sign == right_sign:
        return True
    # <--

    public_key = str(json.dumps(data, ensure_ascii=False, sort_keys=True, separators=(',', ':')))
    public_key = app_id + salt_uuid + public_key + app_key
    right_sign = func_sign_check(public_key)
    return sign == right_sign



__all__ = ('APIHandler', 'APICaller',)

import json
import logging
from abc import ABC
from datetime import datetime, timedelta
from typing import Optional, Awaitable

from ..http import AbstractApiClient, AbstractApiHandler
from ...encrypt import jwt
from ...settings import SETTINGS


class APIHandler(AbstractApiHandler, ABC):
    async def prepare(self, ) -> Optional[Awaitable[None]]:
        access_token = self.request.headers.get('Access-Token')
        if access_token is None:
            return self.finish({
                'msg': 'Invalid Access-Token in request header!'
            })
        else:
            secret_key = SETTINGS.config['APP_SECRET_KEY']
            # options = None
            # TODO: if not validate expiration
            options = {"verify_exp": False}
            access_token = jwt.decode(secret_key, access_token, options=options)
            if not isinstance(access_token, dict):
                self.LOG.error('Invalid Access-Token found in request for [%s]: %s' % (
                    str(self.request.full_url()), access_token
                ))
                return self.finish({
                    'msg': access_token
                })
        self.set_header('Request-ID', self.request_id)

    async def post(self, *args, **kwargs):
        body_arguments = self.request_body
        kwargs.update(body_arguments)
        try:
            if self.LOG.level == logging.DEBUG:
                s_kwargs = json.dumps(kwargs, ensure_ascii=False)
                self.LOG.debug('POST Request [%s]: %s' % (self.request_id, s_kwargs[:1000]))
            self.api_args, self.api_kwargs = args or (), kwargs or {}
            resp = self.response(*self.api_args, **self.api_kwargs)  # this call may throw TypeError when argument missing
        except Exception as e:
            self.LOG.error(e, exc_info=True)
            self.LOG.info('POST Request [%s]: %s' % (self.request_id, self.request.body))
            return self.finish({'status': 'error', 'message': [str(e)]})

        if isinstance(resp, (dict, list)):
            resp = json.dumps(resp, ensure_ascii=False, default=str, separators=(',', ':'))
        return self.finish(resp)

    async def get(self, *args, **kwargs):
        query_arguments = self.request_param
        kwargs.update(query_arguments)
        try:
            self.LOG.debug('GET Request [%s]: %s' % (self.request_id, kwargs))
            self.api_args, self.api_kwargs = args or (), kwargs or {}
            resp = self.response(*self.api_args, **self.api_kwargs)  # this call may throw TypeError when argument missing
        except Exception as e:
            self.LOG.error(e, exc_info=True)
            self.LOG.info('GET Request [%s]: %s' % (self.request_id, kwargs))
            return self.finish({'status': 'error', 'message': [repr(e)]})

        if isinstance(resp, (dict, list)):
            resp = json.dumps(resp, ensure_ascii=False, default=str, separators=(',', ':'))
        return self.finish(resp)


class APICaller(AbstractApiClient):
    APP_ID_KEYS = AbstractApiClient.config.get('APP_ID_KEYS', {})
    APP_SECRET_KEY = AbstractApiClient.config.get('APP_SECRET_KEY')

    def wrap_request_data(self, data: dict) -> dict:
        assert isinstance(data, dict), "Data object must be a dict!"
        return data

    def get_headers(self, app_id: str = None, app_key: str = None) -> dict:
        if app_id is None:
            # if len(APP_ID_KEYS) != 1:
            #     raise RuntimeError('Please specify 1 and only 1 in APP_ID_KEYS in configurations!')
            app_id = list(self.APP_ID_KEYS.keys())[0]

        expire_time = datetime.now() + timedelta(days=1)

        access_token = jwt.encode(
            secret_key=self.APP_SECRET_KEY,
            payload={
                'exp': int(expire_time.timestamp()),
                'aid': app_id
            }
        )

        headers = super().get_headers()
        headers.update({'Access-Token': access_token})
        return headers



__all__ = ('Application',)

import asyncio

from ..logger import LOG

try:
    import uvloop
    from tornado.platform.asyncio import AsyncIOMainLoop

    LOG.info('Using uvloop == %s for service event loop...' % uvloop.__version__)
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    AsyncIOMainLoop().install()
except ImportError:
    LOG.info('[uvloop] NOT installed, fallback to asyncio loop! Consider `pip install uvloop`!')

from .web import WebApplication
from ..settings import SETTINGS

from tornado.options import options


class Application:
    def __init__(self, *args, **kwargs):
        options['log_file_prefix'] = 'access.log'
        settings = dict(SETTINGS.config)
        self.web_app = WebApplication(settings)

    def start(self):
        try:
            self.web_app.start()
            event_loop = asyncio.get_event_loop()
            if event_loop.is_running():
                # notice: the event loop MUST NOT be initialized before web_app starts (as it may fork process)
                # ref: https://github.com/tornadoweb/tornado/issues/2426#issuecomment-400895086
                raise RuntimeError('Event loop already running before WebApp starts!')
            else:
                event_loop.run_forever()
        except KeyboardInterrupt:
            pass
        except Exception as e:
            raise e
        finally:
            pass

    def stop(self):
        event_loop = asyncio.get_event_loop()
        if event_loop.is_running():
            event_loop.stop()



from .config import hocon, paths


class Settings:
    def __init__(self):
        self._config = None

    @property
    def resource_dir(self):
        return paths.get_resource_dir()

    @property
    def config_dir(self):
        return paths.get_config_dir()

    @property
    def config(self):
        if self._config is None:
            config_files = paths.get_config_files()  # by default, use the `main.conf` file in the config_dir
            self._config = hocon.load_config_from_hocon_files(config_files, base_dir=paths.get_config_dir())

        return self._config

    def __getitem__(self, item):
        return self.config[item]


SETTINGS = Settings()



from attrdict import AttrDict
from pyhocon import ConfigFactory


def load_config_from_hocon(config_file):
    config = ConfigFactory.parse_file(config_file).as_plain_ordered_dict()
    return config


def load_config_from_hocon_files(config_files: list, base_dir: str):
    s = []
    for config_file in config_files:
        f = 'include required("%s")' % config_file
        s.append(f)
    f = '\n'.join(s)

    config = ConfigFactory.parse_string(content=f, basedir=base_dir).as_plain_ordered_dict()
    return AttrDict(config)



import sys

from .api import v0, v1, v2
from .http import DefaultHandler404

for module in (v0, v1, v2):
    full_name = '{}.{}'.format(__package__, module.__name__.rsplit('.')[-1])
    sys.modules[full_name] = sys.modules[module.__name__]



import queue
import threading

from .base import BaseStreamer, BaseWorker, TIMEOUT_WORKER, logger


class ThreadedWorker(BaseWorker):
    def __init__(self, func_predict: callable, batch_size: int, max_latency: float,
                 request_queue, response_queue, *args, **kwargs):
        super().__init__(func_predict, batch_size, max_latency, *args, **kwargs)
        self._request_queue = request_queue
        self._response_queue = response_queue


class ThreadedStreamer(BaseStreamer):
    def __init__(self, predict_function: callable, batch_size: int, max_latency: float = 0.1, worker_timeout: float = TIMEOUT_WORKER):
        super().__init__(worker_timeout=worker_timeout)

        self._input_queue = queue.Queue()
        self._output_queue = queue.Queue()

        self._worker_destroy_event = threading.Event()
        self._worker = ThreadedWorker(
            predict_function, batch_size, max_latency, self._input_queue, self._output_queue,
            destroy_event=self._worker_destroy_event
        )
        self._worker_thread = threading.Thread(target=self._worker.run_forever, name="thread_worker")
        self._worker_thread.daemon = True
        self._worker_thread.start()
        self._delay_setup()

    def destroy_workers(self):
        self._worker_destroy_event.set()
        self._worker_thread.join(timeout=self._worker_timeout)
        if self._worker_thread.is_alive():
            raise TimeoutError("worker_thread destroy timeout")
        logger.info("workers destroyed")



import time

import requests

from ...logger import LOG


def iter_over_request_files(request, url_files):
    for file_key, files in request.files.items():  # iter over files uploaded by multipart
        for f in files:
            file_name, content_type = f["filename"], f["content_type"]
            body = f.get('body', b"")
            LOG.info(f"File {file_name} from multipart has content type {content_type} and length bytes={len(body)}")
            yield file_key, file_name, content_type, body

    for file_key, list_url in {'url_files': url_files or []}.items():  # iter over files specified by `url_files`
        for url in sorted(set(list_url)):
            try:
                t_start = time.time()
                resp = requests.get(url, stream=True)  # download the file from given url
                if resp.status_code == 200:
                    body = resp.content
                    content_type = resp.headers.get("Content-Type", "UNKNOWN")
                else:
                    raise RuntimeError("Failed to download file after %s seconds with code=%s from URL %s" % (
                        time.time() - t_start, resp.status_code, url
                    ))
                del resp
            except Exception as e:
                raise e
            t_cost = time.time() - t_start
            LOG.info(f"File {url} has content type {content_type} and length bytes={len(body)}, downloaded in {t_cost} seconds")
            yield 'url_files', url, content_type, body



__all__ = ('APIHandler', 'APICaller',)

import json
import logging
from abc import ABC

from ..http import AbstractApiClient, AbstractApiHandler


class APIHandler(AbstractApiHandler, ABC):
    MAP_ERROR_INFO = {
        'BAD_REQUEST': {'code': '5101', 'message': ['Bad request: fail to parse body as JSON object!']}
    }

    async def post(self, *args, **kwargs):
        req_body = self.request_body

        if req_body is not None:  # body_arguments
            kwargs.update(req_body)

        resp = dict(code=5200, message=['success'])
        try:
            result = self.response(*args, **kwargs)  # this call may throw TypeError when argument missing
            resp['data'] = result
        except Exception as e:
            if self.LOG.level == logging.DEBUG:
                self.LOG.error(e, exc_info=True)
            return self.finish({'code': 5201, 'message': [repr(e)]})

        resp = json.dumps(resp, ensure_ascii=False, default=str, separators=(',', ':'))
        return self.finish(resp)


class APICaller(AbstractApiClient):
    def wrap_request_data(self, data: dict) -> dict:
        assert isinstance(data, dict), "Data object must be a dict!"
        return data



from ..encrypt import vault
from ..logger import LOG
from ..settings import SETTINGS


class PasswordVault:
    _dict_cache_vault = {}

    @staticmethod
    def get_vault(vault_type: str = None, vault_config: dict = None, *args, **kwargs) -> vault.BaseVault:
        encryption_method = vault_type or SETTINGS.config.get('PASSWORD_ENCRYPTION')
        LOG.debug('Using password vault: %s', encryption_method)

        cache_key = '%s:%s' % (encryption_method, str(vault_config))
        if cache_key not in PasswordVault._dict_cache_vault:
            if encryption_method in ('plain', 'aes') or encryption_method is True:
                v = vault.AesVault(**(vault_config or {}))
            elif encryption_method == 'cyberark':
                config_cyberark = vault_config or SETTINGS.config.get('CYBERARK_CONFIG')
                if config_cyberark is None:
                    raise RuntimeError('Missing [CYBERARK_CONFIG] in config!')
                v = vault.CyberArkVault(**config_cyberark)
            else:
                LOG.info('Using plain password vault as unknown value of PASSWORD_ENCRYPTION=%s in config.', encryption_method)
                v = vault.DummyVault(**(vault_config or {}))
            PasswordVault._dict_cache_vault[cache_key] = v

        return PasswordVault._dict_cache_vault[cache_key]


def main():
    import sys
    config_key = sys.argv[-1]
    LOG.debug('Getting pwd for deploy key [deploy.%s]' % config_key)
    try:
        db_config = SETTINGS.config['deploy'][config_key]
        password_vault = PasswordVault.get_vault()
        p = password_vault.get_password(db_config.get('password'))
        LOG.debug("Decrypted PWD: %s" % p)
    except KeyError:
        LOG.error('Please make sure config key [deploy.%s] exists!' % config_key)



import uuid
from abc import ABC, abstractmethod
from urllib.parse import urljoin

import requests
from requests.adapters import HTTPAdapter, Retry

from ...logger import LOG
from ...settings import SETTINGS


class AbstractApiClient(ABC):
    LOG = LOG
    RETRY_METHOD_WHITELIST: frozenset = frozenset(['GET', 'POST'])
    RETRY_STATUS_FORCELIST: frozenset = frozenset({413, 429, 503, 502, 504})
    config = SETTINGS.config

    def __init__(self, url_endpoint: str = None, *args, **kwargs):
        self.url_endpoint = url_endpoint or ''
        LOG.debug('API Caller URL endpoint set to: %s' % self.url_endpoint)

    @classmethod
    def get_request_session(cls, total_retries: int = 3, *args, **kwargs) -> requests.Session:
        session = requests.Session()
        # https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.Retry.DEFAULT_ALLOWED_METHODS
        retries = Retry(
            total=total_retries, backoff_factor=0.1, method_whitelist=cls.RETRY_METHOD_WHITELIST, status_forcelist=cls.RETRY_STATUS_FORCELIST
        )
        for prefix in ('http://', 'https://'):
            session.mount(prefix, HTTPAdapter(max_retries=retries))
        return session

    def get_headers(self, *args, **kwargs) -> dict:
        headers = {
            'Content-Type': 'application/json',
            'Request-ID': str(uuid.uuid1()),
        }
        return headers

    @abstractmethod
    def wrap_request_data(self, data: dict) -> dict:
        assert isinstance(data, dict), "Data object must be a dict!"
        raise NotImplementedError()
        # return data

    def call(self, api_url: str, data: dict = None, timeout=5, **kwargs):
        """Trigger API call
        :param api_url: do NOT start with slash (/)
        :param data: a dictionary which includes the request data
        :param timeout: requests timeout in seconds
        :param kwargs: keywords arguments which will be updated to data
        :return:
        """
        body = data or dict()
        body.update(kwargs)
        payload = self.wrap_request_data(data=body)
        LOG.debug('Calling api: %s' % api_url)
        session = self.get_request_session()
        resp = session.post(
            urljoin(self.url_endpoint, api_url), json=payload, timeout=timeout, headers=self.get_headers()
        )

        try:
            ret = resp.json()
        except Exception as e:
            LOG.error(str(e))
            raise RuntimeError(resp.text)

        return ret



from .logger import get_logger
from ..settings import SETTINGS

LOG = get_logger(
    level=SETTINGS.config.get('deploy', {}).get('log_level', 10),  # 10 = logging.DEBUG
    # logger_name='default',
)



__all__ = ('get_resource_dir', 'get_config_dir', 'get_current_module_dir', 'get_project_base_dir', 'path_join')

import os
import sys
import warnings


def path_join(*args) -> str:
    p = os.path.join(*args)
    p = os.path.expanduser(p)
    p = os.path.expandvars(p)
    p = os.path.abspath(p)
    return p


def get_resource_dir(*args) -> str:
    dir_cwd = os.getcwd()
    dir_resource = os.environ.get('DIR_RESOURCE', 'resource')
    return path_join(dir_cwd, dir_resource, *args)


def get_config_dir(*args) -> str:
    dir_config = os.environ.get('DIR_CONFIG')
    dir_resource = get_resource_dir()
    if dir_config is None or len(dir_config.strip()) == 0:
        dir_config = 'config'
    dir_config = path_join(dir_resource, dir_config, *args)
    return dir_config


def get_config_files() -> list:
    """
    Get a list of config files to parse. The base dir of config files should be specified by `get_config_dir()`.
    1. The function will look up the `FILES_CONFIG` environment variable to get a list of file names seperated by comma, if specified;
    2. In case `FILES_CONFIG` is not specified, the function will use the default config file;
    3. The default config file is determined by:
       (a) If environment variable `ENV_PROFILE` is defined, the entry file will be "main-{ENV_PROFILE}.conf"
       (b) If environment variable `ENV_PROFILE` is not defined, the entry config file will be "main.conf".
    :return: list of string, which are file names of config files
    """
    files_config = os.environ.get('FILES_CONFIG', None)
    if files_config is None:
        env_profile = os.environ.get('ENV_PROFILE', None)
        if env_profile is None:
            files_config = 'main.conf'
        else:
            files_config = 'main-%s.conf' % env_profile

    files = files_config.split(',')
    ret = []
    msgs = []
    for f in files:
        file = get_config_dir(f)
        if not os.path.exists(file):
            msgs.append('Expecting config file [%s] but it does not exists!' % file)
        else:
            print('  ---> Loading config file [%s]' % file, file=sys.stderr)
            ret.append(os.path.expandvars(f))
    if len(ret) == 0:
        msgs.append('No config files set properly, EMPTY config will be used!')

    if len(msgs) > 0:
        warnings.warn('\n'.join(msgs))
    return ret


def get_current_module_dir(file_caller: str) -> str:
    dirs = file_caller.split(os.sep)
    return os.sep.join(dirs[:-1])


def get_project_base_dir(file_caller: str) -> str:
    dirs = file_caller.split(os.sep)

    dir_base = ''
    for i in range(len(dirs)):
        dir_base = os.sep.join(dirs[:-1 - i])
        file_init = path_join(dir_base, '__init__.py')
        if not os.path.exists(file_init):
            break
    return dir_base



import json
from abc import ABC
from datetime import datetime
from typing import Optional, Awaitable

from tornado import web

from ...logger import LOG


class AbstractApiHandler(web.RequestHandler, ABC):
    LOG = LOG
    MAP_ERROR_INFO: dict = {
        'BAD_REQUEST': {'code': '5101', 'message': ['Bad request: fail to parse body as JSON object!']}
    }

    def __init__(self, *args, **kwargs):
        self.api_args: Optional[tuple] = None
        self.api_kwargs: Optional[dict] = None
        super().__init__(*args, **kwargs)

    def on_finish(self) -> None:
        func_callback = getattr(self, 'callback', None)
        if callable(func_callback) \
                and isinstance(self.api_args, tuple) \
                and isinstance(self.api_kwargs, dict):
            func_callback(*self.api_args, **self.api_kwargs)

    def response(self, *args, **kwargs) -> dict:
        raise NotImplementedError()

    def set_default_headers(self) -> None:
        self.set_header('Content-Type', 'application/json; charset=utf-8')

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    @property
    def request_header_content_type(self) -> str:
        return self.request.headers.get('Content-Type', 'application/json; charset=utf-8')

    @property
    def request_id(self):
        if 'Request-ID' not in self.request.headers:
            self.request.headers['Request-ID'] = datetime.now().strftime('%Y%m%d-%H%M%S-%f')
        return self.request.headers.get('Request-ID')

    @property
    def request_body(self) -> dict:
        content_type: str = self.request_header_content_type
        body_arguments: dict = Optional[None]

        if content_type.startswith('multipart/form-data'):  # only parse files when 'Content-Type' starts with 'multipart/form-data'
            body_arguments = self.request_param  # self.request.body_arguments
        else:
            try:
                body = self.request.body.decode('utf-8')
                body_arguments = json.loads(body)
            except (UnicodeDecodeError, json.decoder.JSONDecodeError):  # invalid request body, cannot be parsed as JSON
                self.finish(self.MAP_ERROR_INFO['BAD_REQUEST'])
        return body_arguments

    @property
    def request_param(self) -> dict:
        ret: dict = {}
        for k, v in self.request.arguments.items():
            val = v[0].decode('utf-8')
            try:
                value = json.loads(val)
            except json.JSONDecodeError:
                value = val
            ret[k] = value

        return ret


class DefaultHandler404(AbstractApiHandler):
    def response(self, *args, **kwargs) -> Optional[dict]:
        return self.prepare()

    def prepare(self):  # for all methods
        msg = {
            "code": 404,
            "status": "error",
            "message": [
                'Requested URL cannot be found: %s' % self.request.uri
            ]
        }
        msg = json.dumps(msg, ensure_ascii=False, default=str, separators=(',', ':'))
        self.set_status(404, reason='Not Found')
        self.finish(msg)



import importlib
import os
import sys


def main():
    print('\n'.join(sorted('%s=%s' % (k, v) for k, v in os.environ.items())))

    usage = '''
    Usage: `python main.py app_common.main` ; or set environment variable `ENTRYPOINT` 
        the `module.name` should be a python package file or package which include a `main()` function
    '''

    try:
        module_name = sys.argv[1]
    except IndexError:
        module_name = os.environ.get('ENTRYPOINT')

    if module_name is None:
        print(usage)
        exit(-1)

    try:
        m = importlib.import_module(module_name)
    except ImportError:
        raise ValueError('Invalid entrypoint: %s' % module_name)

    f_main = getattr(m, 'main')

    if f_main is None:
        print('Given module does not provides a `main()` function!')
    else:
        print('Starting module: %s' % module_name)
        ret = f_main()
        if ret:
            print(ret)


if __name__ == '__main__':
    """aloha start module.name"""
    """python -m aloha.script.start module.name"""
    main()



__all__ = ('SqliteOperator',)

import sqlite3

from sqlalchemy import create_engine
from sqlalchemy.sql import text

from .base import PasswordVault
from ..logger import LOG


class SqliteOperator:
    def __init__(self, db_config, **kwargs):
        self._connection_pattern = "sqlite://{dbname}"
        dbname = db_config.get('dbname', '')
        if len(dbname) > 0:
            dbname = '/%s' % dbname
        self._config = {'dbname': dbname}

        if 'password' in db_config:
            try:
                import sqlcipher3
            except ImportError:
                raise RuntimeError('Python package required for encrypted sqlite3: sqlcipher3-binary')
            LOG.debug('Version of sqlcipher3 = %s' % sqlcipher3.sqlite_version)
            password_vault = PasswordVault.get_vault(db_config.get('vault_type'), db_config.get('vault_config'))
            password = password_vault.get_password(db_config.get('password', None))
            self._config['password'] = password
            self._connection_pattern = "sqlite+pysqlcipher://:{password}@/{dbname}"
        else:
            LOG.debug('Version of sqlite = %s' % sqlite3.sqlite_version)

        try:
            self.db = create_engine(
                self._connection_pattern.format(**self._config), **kwargs
            )
            LOG.debug("Sqlite connected: %s" % self.connection_str)
        except Exception as e:
            LOG.exception(e)
            raise RuntimeError('Failed to connect to sqlite')

    @property
    def connection(self):
        return self.db

    def execute_query(self, sql, *args, **kwargs):
        with self.db.connect() as conn:
            cur = conn.execute(text(sql), *args, **kwargs)
            return cur

    @property
    def connection_str(self) -> str:
        return self._connection_pattern.format(**self._config)
